from marshmallow import (
    Schema,
    fields,
    ValidationError,
    validates_schema,
    INCLUDE,
    validate,
)
from .choices import (
    CHANNELS,
    SEND_TO,
    COMMUNICATION_LEVEL,
    BATCH_TYPE,
    BATCH_CHANNELS,
    APPLICANT_TYPE,
    REMINDER_OFFSET_TYPE,
    NOTIFY_TYPE,
    RESULT_CODE,
    NOTIFICATION_SOURCE,
    TypeOfTaskChoices,
)


class CampaignDataInput(Schema):
    class Meta:
        unknown = INCLUDE

    name = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="name should not be empty string"),
        error_messages={"required": "name data is required"},
    )
    channel = fields.Str(
        required=True,
        validate=validate.OneOf(CHANNELS),
        error_messages={"required": "channel is required"},
    )
    type = fields.Str(
        required=True,
        validate=validate.OneOf(["communication"]),
        error_messages={"required": "channel is required"},
    )
    description = fields.Str()


class CampaignInput(Schema):
    class Meta:
        unknown = INCLUDE

    campaign_data = fields.Nested(
        CampaignDataInput,
        required=True,
        error_messages={"required": "campaign data is required"},
    )
    loan_data = fields.List(
        fields.String(),
        required=True,
        validate=validate.Length(min=1, error="loan data should not be empty"),
        error_messages={"required": "loan data is required"},
    )
    comm_dict = fields.Dict()
    additional_loan_data = fields.Dict()
    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    template_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="template id should not be empty string"),
        error_messages={"required": "template id is required"},
    )
    # allocation_month = fields.DateTime(
    #     required=True,
    #     format='%Y-%m-%d',
    #     error_messages={"required": "Allocation month is required"}
    # )
    allocation_month = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="allocation month should not be empty"),
        error_messages={"required": "Allocation month is required"},
    )
    upto_index = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="upto index should not be empty string"),
        error_messages={"required": "upto index is required"},
    )
    send_to = fields.Str(
        required=True,
        validate=[
            validate.OneOf(SEND_TO),
            validate.Length(min=1, error="send to should not be empty string"),
        ],
        error_messages={"required": "send to is required"},
    )
    communication_level = fields.Str(
        required=True,
        validate=validate.OneOf(COMMUNICATION_LEVEL),
        error_messages={"required": "communication level is required"},
    )

    @validates_schema
    def validate_data(self, data, **kwargs):
        other_errors = []
        campaign_data = data.get("campaign_data", None)
        comm_dict = data.get("comm_dict", None)
        if comm_dict in (None, {}) and campaign_data["channel"] != "dtmf_ivr":
            other_errors.append("Comm dict is required")
        if other_errors:
            raise ValidationError({"message": other_errors})


class APIResponseInput(Schema):
    class Meta:
        unknown = INCLUDE

    status = fields.Str(required=True)
    response = fields.Str(required=True)
    loan_id = fields.Str(required=True)
    status_code = fields.Str(required=True)
    tracking_number = fields.Str()
    company_id = fields.UUID()
    campaign_id = fields.Str()


class GetBatchData(Schema):
    class Meta:
        unkown = INCLUDE

    batch_id = fields.List(
        fields.String(),
        validate=validate.Length(min=1, error="Please provide valid value"),
        required=True,
    )
    company_id = fields.UUID()


class UpdateRecordData(Schema):
    class Meta:
        unknown = INCLUDE

    table = fields.Str(required=True)
    set_clause = fields.Dict(required=True)
    from_values = fields.List(
        fields.String(),
        validate=validate.Length(min=1, error="Please provide valid value"),
        required=True,
    )
    from_columns = fields.List(
        fields.String(),
        validate=validate.Length(min=2, error="Please provide valid value for columns"),
        required=True,
    )
    from_alias_name = fields.Str(required=True)
    where = fields.Dict(required=True)


class ResponseInput(Schema):
    class Meta:
        unknown = INCLUDE

    payload = fields.Dict(required=True, error_messages={"required": "payload is required"})
    response = fields.Nested(
        APIResponseInput,
        required=True,
        error_messages={"required": "response is required"},
    )

    @validates_schema
    def validate_data(self, data, **kwargs):
        other_errors = []
        payload = data.get("payload")
        response = data.get("response")
        if payload.get("type", "") == TypeOfTaskChoices.scrape.value and (not "tracking_number" in response):
            other_errors.append("Please provide valid tracking number")
        if payload.get("type", "") == TypeOfTaskChoices.scrape.value and (not "company_id" in response):
            other_errors.append("Please provide valid company id")
        if other_errors:
            raise ValidationError({"error": other_errors})


class ResponseBatchInput(Schema):
    class Meta:
        unknown = INCLUDE

    payload = fields.Dict(required=True, error_messages={"required": "payload is required"})
    insert_list = fields.List(
        fields.Nested(APIResponseInput),
        required=True,
        error_messages={"required": "insert list is required"},
    )


class ReportInput(Schema):
    class Meta:
        unknown = INCLUDE

    campaign_id = fields.UUID(required=True, error_messages={"required": "campaign id is required"})
    channel = fields.Str(
        required=True,
        validate=validate.OneOf(CHANNELS),
        error_messages={"required": "channel is required"},
    )
    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})


class GetCampaignInput(Schema):
    class Meta:
        unknown = INCLUDE

    campaign_id = fields.UUID(required=True, error_messages={"required": "campaign id is required"})
    channel = fields.Str(
        required=True,
        validate=[
            validate.OneOf(CHANNELS),
            validate.Length(min=1, error="channel should not be empty string"),
        ],
        error_messages={"required": "channel is required"},
    )

    @validates_schema
    def validate_data(self, data, **kwargs):
        other_errors = []
        if not data.get("channel", None) or data.get("channel", "") not in CHANNELS:
            other_errors.append("Please select valid channel")
        if other_errors:
            raise ValidationError({"message": other_errors})


class GetBatchInput(Schema):
    class Meta:
        unknown = INCLUDE

    batch_id = fields.UUID(required=True, error_messages={"required": "campaign id is required"})
    channel = fields.Str(required=True, error_messages={"required": "channel is required"})
    type = fields.Str(
        required=True,
        validate=[
            validate.OneOf(BATCH_TYPE),
            validate.Length(min=1, error="type should not be empty string"),
        ],
        error_messages={"required": "type is required"},
    )

    @validates_schema
    def validate_data(self, data, **kwargs):
        other_errors = []
        channel = data.get("channel", "")
        if channel and channel not in (BATCH_CHANNELS):
            other_errors.append(f"Please provide valid channel")


class BatchOperationInput(Schema):
    class Meta:
        unknown = INCLUDE

    type = fields.Str(
        required=True,
        validate=validate.OneOf(BATCH_TYPE, error="Please provide valid type"),
        error_messages={"required": "type is required"},
    )
    channel = fields.Str(
        # validate=validate.OneOf(BATCH_CHANNELS, error = 'Please provide valid channel'),
        # error_messages={"required": "channel is required"}
    )

    @validates_schema
    def validate_data(self, data, **kwargs):
        other_errors = []
        type = data.get("type", "")
        channel = data.get("channel", "")
        is_linked_loan = data.get("is_linked_loan", False)
        if channel and channel not in (BATCH_CHANNELS):
            other_errors.append(f"Please provide valid channel")
        elif type == TypeOfTaskChoices.scrape.value and channel not in (
            TypeOfTaskChoices.indiapost_upload.value,
            TypeOfTaskChoices.indiapost_tracking.value,
        ):
            other_errors.append(f"Please provide valid channel")
        elif type == TypeOfTaskChoices.notice.value and (
            channel
            not in (
                TypeOfTaskChoices.linked_loan_digital.value,
                TypeOfTaskChoices.digital.value,
                TypeOfTaskChoices.physical.value,
                TypeOfTaskChoices.linked_loan_physical.value,
            )
        ):
            other_errors.append(f"Please provide valid channel")
        elif (
            type
            not in (
                TypeOfTaskChoices.notice.value,
                TypeOfTaskChoices.scrape.value,
                TypeOfTaskChoices.ecourt_tracking.value,
            )
        ) and (channel or data.get("channel", None) == ""):
            other_errors.append(f"channel is not required for type:{type}")
        if type == TypeOfTaskChoices.scrape.value and channel == TypeOfTaskChoices.indiapost_upload.value:
            result = IndiapostUploadQueue().validate(data)
            if result != {}:
                other_errors.append(result)
        if type == TypeOfTaskChoices.scrape.value and channel == TypeOfTaskChoices.indiapost_tracking.value:
            result = IndiapostTrackingQueue().validate(data)
            if result != {}:
                other_errors.append(result)
        if type == TypeOfTaskChoices.notice.value and channel == TypeOfTaskChoices.digital.value:
            if is_linked_loan:
                result = LinkedLoanQueue().validate(data)
            else:
                result = DigitalNoticeQueue().validate(data)
            if result != {}:
                other_errors.append(result)
        if type == TypeOfTaskChoices.lat_long_conversion.value:
            result = LatLongConversionQueue().validate(data)
            if result != {}:
                other_errors.append(result)
        if type == TypeOfTaskChoices.upload_c2c_disposition.value:
            result = UploadDispositionQueue().validate(data)
            if result != {}:
                other_errors.append(result)
        if type == TypeOfTaskChoices.optin.value:
            result = OptinQueue().validate(data)
            if result != {}:
                other_errors.append(result)
        if type == TypeOfTaskChoices.notifications.value:
            result = NotificationsQueue().validate(data)
            if result != {}:
                other_errors.append(result)
            if data.get("notify_type", "") == "reminder" and (not data.get("reminder_offset")):
                other_errors.append(f"reminder offset is required for notify_type: reminder")
            if data.get("notify_type", "") == "reminder" and (not data.get("reminder_offset_type")):
                other_errors.append(f"reminder offset type is required for notify type: reminder")
        if other_errors:
            raise ValidationError({"error": other_errors})


class NotificationsData(Schema):
    class Meta:
        unknown = INCLUDE

    message = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid message"),
        error_messages={"required": "message is required"},
    )


class NotificationsQueue(Schema):
    class Meta:
        unknown = INCLUDE

    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    source = fields.Str(
        required=True,
        validate=validate.OneOf(NOTIFICATION_SOURCE, error="Please provide valid source"),
        error_messages={"required": "source is required"},
    )
    notify_type = fields.Str(
        required=True,
        validate=validate.OneOf(NOTIFY_TYPE, error="Please provide valid notify type"),
        error_messages={"required": "notify type is required"},
    )
    reminder_offset_type = fields.Str(
        validate=validate.OneOf(REMINDER_OFFSET_TYPE, error="Please provide valid reminder offset type")
    )
    trigger_time = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid trigger time"),
        error_messages={"required": "trigger time is required"},
    )
    reminder_offset = fields.List(fields.Integer(strict=True))
    loan_data = fields.List(
        fields.Nested(NotificationsData),
        required=True,
        error_messages={"required": "Loan data is required"},
    )


class OptinQueue(Schema):
    class Meta:
        unknown = INCLUDE

    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    loan_data = fields.List(
        fields.String(),
        validate=validate.Length(min=1, error="No loan ids provided"),
        required=True,
        error_messages={"required": "loan data is required"},
    )


class UploadDispositionData(Schema):
    class Meta:
        unknown = INCLUDE

    loan_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid loan id"),
        error_messages={"required": "loan id is required"},
    )
    agent_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid agent id"),
        error_messages={"required": "agent id is required"},
    )
    rec_url = fields.Str(required=True, error_messages={"required": "rec_url is required"})
    to_mobile = fields.Str(required=True, error_messages={"required": "to_mobile key is required"})
    call_duration = fields.Str(required=True, error_messages={"required": "call duration is required"})
    call_type = fields.Str()
    call_status = fields.Str()
    call_start_time = fields.Str(required=True, error_messages={"required": "call start time is required"})
    call_end_time = fields.Str(required=True, error_messages={"required": "call end time is required"})
    campaign_id = fields.Str(
        required=True,
        allow_none=True,
        error_messages={"required": "campaign id is required"},
    )
    result_code = fields.Str(
        required=True,
        validate=validate.OneOf(RESULT_CODE, error="Please provide valid result code"),
        error_messages={"required": "result code is required"},
    )


class UploadDispositionQueue(Schema):
    class Meta:
        unknown = INCLUDE

    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    loan_data = fields.List(
        fields.Nested(UploadDispositionData),
        required=True,
        error_messages={"required": "Loan data is required"},
    )


class LatLongConversionData(Schema):
    class Meta:
        unknown = INCLUDE

    loan_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid loan id"),
        error_messages={"required": "loan id is required"},
    )
    address = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid address"),
        error_messages={"required": "address is required"},
    )
    applicant_index = fields.Integer(
        strict=True,
        required=True,
        allow_none=True,
        error_messages={"required": "applicant_index is required"},
    )
    address_index = fields.Integer(
        strict=True,
        required=True,
        error_messages={"required": "address_index is required"},
    )


class LatLongConversionQueue(Schema):
    class Meta:
        unknown = INCLUDE

    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    applicant_type = fields.Str(
        required=True,
        validate=validate.OneOf(APPLICANT_TYPE, error="Please provide valid applicant_type"),
        error_messages={"required": "applicant_type is required"},
    )
    loan_data = fields.List(
        fields.Nested(LatLongConversionData),
        required=True,
        error_messages={"required": "Loan data is required"},
    )


class LinkedLoanDataInput(Schema):
    class Meta:
        unknown = INCLUDE

    linked_loan_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid linked_loan_id"),
        error_messages={"required": "linked_loan_id is required"},
    )
    loan_ids = fields.List(
        fields.String(),
        validate=validate.Length(min=1, error="No loan ids provided"),
        required=True,
        error_messages={"required": "loan ids is required"},
    )


class LinkedLoanQueue(Schema):
    class Meta:
        unknown = INCLUDE

    draft_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid draft id"),
        error_messages={"required": "Please provide Draft id"},
    )
    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    creditline = fields.Str()
    delivery_partner = fields.Str()
    allocation_month = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid allocation month"),
        error_messages={"required": "Allocation month is required"},
    )
    all_applicants = fields.Str()
    co_applicant = fields.Str()
    loan_data = fields.List(
        fields.Nested(LinkedLoanDataInput),
        required=True,
        error_messages={"required": "Loan data is required"},
    )


class DigitalNoticeQueue(Schema):
    class Meta:
        unknown = INCLUDE

    draft_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid draft id"),
        error_messages={"required": "Please provide Draft id"},
    )
    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    creditline = fields.Str()
    delivery_partner = fields.Str()
    allocation_month = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid allocation month"),
        error_messages={"required": "Allocation month is required"},
    )
    all_applicants = fields.Str()
    co_applicant = fields.Str()
    loan_data = fields.List(
        fields.String(),
        validate=validate.Length(min=1, error="No loan ids provided"),
        required=True,
        error_messages={"required": "loan data is required"},
    )


class IndiapostUploadData(Schema):
    class Meta:
        unknown = INCLUDE

    loan_id = fields.Str(validate=validate.Length(min=1, error="Please provide valid loan id"))
    customer_name = fields.Str()
    customer_address1 = fields.Str()
    customer_address2 = fields.Str()
    customer_address3 = fields.Str()
    customer_city = fields.Str()
    pin_code = fields.Str()
    customer_state = fields.Str()
    notice_id = fields.Str()
    tracking_no = fields.Str()


class IndiapostUploadQueue(Schema):
    class Meta:
        unknown = INCLUDE

    notice_type = fields.Str()
    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    allocation_month = fields.Str()
    loan_data = fields.List(fields.Nested(IndiapostUploadData))


class IndiapostTrackingData(Schema):
    class Meta:
        unknown = INCLUDE

    loan_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid loan id"),
        error_messages={"required": "loan id is required"},
    )
    company_id = fields.UUID(required=True, error_messages={"required": "company id is required"})
    tracking_id = fields.Str(
        required=True,
        validate=validate.Length(min=1, error="Please provide valid tracking id"),
        error_messages={"required": "tracking id is required"},
    )


class IndiapostTrackingQueue(Schema):
    class Meta:
        unknown = INCLUDE

    loan_data = fields.List(fields.Nested(IndiapostTrackingData))



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
import pickle
import os
import uuid


def get_required_headers_per_level(headers):
    logger.info("get_required_headers_per_level")

    loan_level_headers = []
    case_level_headers = []
    stage_level_headers = []
    reminder_level_headers = []
    advocate_level_headers = []

    for key, value in headers.items():
        level = value.get("level")
        if not level:
            pass
        elif level == "loan":
            loan_level_headers.append(key)
        elif level == "stage":
            stage_level_headers.append(key)
        elif level == "case":
            case_level_headers.append(key)
        elif level == "reminder":
            reminder_level_headers.append(key)
        elif level == "advocate":
            advocate_level_headers.append(key)

    # adding necessary keys - various ids in all dicts -
    for header_array in (
        loan_level_headers,
        case_level_headers,
        stage_level_headers,
        reminder_level_headers,
        advocate_level_headers,
    ):
        for key in [
            "id",
            "case_row_id",
            "loan_id",
            "case_id",
            "stage_code",
            "current_stage_code",
            "proceeding",
            "data",
        ]:
            if key not in header_array:
                header_array.append(key)

    # adding specific keys -
    stage_level_headers.append("case_details_under_approval")
    stage_level_headers.append("iteration")
    reminder_level_headers.append("reminder_start_date")
    reminder_level_headers.append("deadline_date")
    reminder_level_headers.append("date_of_default")
    reminder_level_headers.append("next_stage_code")
    advocate_level_headers.append("user_id")
    advocate_level_headers.append("profession")
    advocate_level_headers.append("iteration")

    return (loan_level_headers, case_level_headers, stage_level_headers, reminder_level_headers, advocate_level_headers)


class ChunkClass:
    file_list = []
    batch_id = None
    counter = -1
    extension = None
    base_path = None
    chunks_file_path = None
    excel_file_path = None
    zip_file_path = None

    def __init__(self, base_path, batch_id, extension) -> None:
        self.file_list = []
        self.batch_id = batch_id
        self.counter = 0
        self.extension = extension
        os.makedirs(base_path, exist_ok=True)

        self.base_path = os.path.join(base_path, batch_id)
        os.makedirs(self.base_path, exist_ok=True)

        self.excel_file_path = f"{os.path.join(self.base_path,'excel')}"
        os.makedirs(self.excel_file_path, exist_ok=True)

        self.zip_file_path = f"{os.path.join(self.base_path,'zip')}"
        os.makedirs(self.zip_file_path, exist_ok=True)

        self.chunks_file_path = f"{os.path.join(self.base_path,'chunks')}"
        os.makedirs(self.chunks_file_path, exist_ok=True)

    def load_data(self):
        logger.info("ChunkClass.load_data")
        filename = self.file_list[self.counter]
        filename = os.path.join(self.chunks_file_path, filename)
        try:
            with open(filename, "rb") as f:
                data_load = pickle.load(f)
                return data_load

        except Exception as e:
            logger.error(f"Exception while loading pickle file - : {str(e)}")
            raise e

    def save_data(self, data):
        logger.info("ChunkClass.save_data")
        if data is None or len(data) == 0:
            logger.debug(f"ChunkClass.save_data || file not saved - 0 data")
            return

        filename = f"{str(uuid.uuid4())}{self.extension}"
        logger.debug(f"ChunkClass.save_data || {filename} len : {len(data)}")
        self.counter += 1
        self.file_list.append(filename)
        filename = os.path.join(self.chunks_file_path, filename)
        try:
            with open(filename, "wb") as file:
                pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)
        except Exception as e:
            logger.error(f"Exception while saving pickle file - : {str(e)}")
            raise e

    def pop_data(self):
        logger.info("ChunkClass.pop_data")
        if self.counter <= len(self.file_list):
            self.counter -= 1
            return self.load_data()
        logger.info("ChunkClass.pop_data failed because all files are loaded")
        return False

    def create_file_list(self):
        logger.info("ChunkClass.create_file_list")
        self.file_list = os.listdir(self.chunks_file_path)
        self.counter = len(self.file_list)
        logger.info(f"ChunkClass.create_file_list || {self.batch_id} || len of file_list : {len(self.file_list)}")


class LitigationReportRedisKeys:
    batch_count = None
    batch_kwargs = None
    callback_kwargs = None
    task_state = None
    batch_number = None
    headers = None

    def __init__(self, batch_id) -> None:
        self.batch_count = f"{batch_id}_loan_ids_batches"
        self.batch_kwargs = f"{batch_id}_batch_kwargs"
        self.callback_kwargs = f"{batch_id}_callback_kwargs"
        self.task_state = f"{batch_id}_report_state"
        self.headers = f"{batch_id}_headers"

    def set_batch_number_key(self, batch_number):
        self.batch_number = batch_number
        self.headers = f"{self.headers}_{self.batch_number}"



# import json
# from app.settings import NOTICE_SERVICE_BASE_URL, REDIS, NOTICE_COMPANIES_WITH_PROPERTY_ADDRESS_HANDLING
# import os
# import copy
# import random
# from datetime import datetime
# from num2words import num2words
# import logging
# import re
# from dateutil.parser import parse
# import string
# from app.tasks.notice.loan_service import get_borrower_language
# from app.utils import get_s3_client
# import requests
# import redis

# logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# s3 = get_s3_client()


# def get_notice_generation_data(**kwargs):
#     logger.info(f"tasks.notice.get_notice_generation_data.batch_id :: {kwargs['batch_id']}")

#     company_id = kwargs.get("company_id", "")
#     draft_id = kwargs.get("draft_id", "")
#     loan_ids = kwargs.get("loan_ids", "")
#     allocation_month = kwargs.get("allocation_month", "")
#     batch_number = kwargs.get("batch_number", "")
#     document_type = kwargs.get("document_type")
#     case_id = kwargs.get("case_id", "")
#     iteration = kwargs.get("iteration", "null")
#     stage_code = kwargs.get("stage_code", "null")
#     linked_loan_id = kwargs.get("linked_loan_id", "")

#     logger.debug(f"notice.get_notice_generation_data.kwargs : {kwargs}")

#     try:
#         master_table_dict = {}
#         table_dict_list = []
#         changed_dt_list = []
#         address_set = set()
#         address_tables = set()

#         loan_data = kwargs["loan_data"]

#         if loan_data:
#             for loan in loan_data:
#                 loan_id = loan.get("loan_id", "")
#                 question_dict = loan.get("question_dict", "")
#                 languages = get_borrower_language(question_dict)
#                 languages["draft_language"] = kwargs.get("language")
#                 allocation_month = question_dict.get("allocation_month", "")
#                 kwargs["allocation_month"] = allocation_month

#                 # axis_bank condition
#                 if company_id == "d2d50f3e-4b79-40d7-b9dd-6cadb0b30141" and question_dict["loan_type"] == "Credit Card":
#                     question_dict["loan_id"] = question_dict.get("credit_card_number", "")
#                     loan_id = question_dict.get("credit_card_number", "")

#                 kwargs["loan_id"] = loan_id
#                 kwargs["question_dict"] = question_dict
#                 para_table_list = get_para_table_dicts(address_set, address_tables, **kwargs)
#                 if isinstance(para_table_list, str):
#                     return para_table_list
#                 changed_dt, table_dict = para_table_list[0], para_table_list[1]
#                 table_dict_list.append(table_dict)
#                 changed_dt_list.append(changed_dt)
#         else:
#             logger.error("get_notice_generation_data.exception: Cannot find data for the given loan")
#             raise Exception("Cannot find data for the given loan")

#         for i in range(len(table_dict_list)):
#             for key, val in table_dict_list[i].items():
#                 if key not in master_table_dict:
#                     master_table_dict[key] = table_dict_list[i][key]
#                 else:
#                     master_table_dict[key].extend(table_dict_list[i][key])

#         is_master_table_dict_empty = True

#         for key, val in master_table_dict.items():
#             if len(val) != 0:
#                 is_master_table_dict_empty = False

#         master_table_dict_list = duplicate_table_populator(master_table_dict, kwargs["draft_fields"])
#         master_table_dict = master_table_dict_list[0]
#         duplicated_tables_dict = master_table_dict_list[1]

#         if kwargs["mode"] == "physical" and is_master_table_dict_empty:
#             logger.error(
#                 "get_notice_generation_data.exception: Either draft does not have address tables or loan doesn't have addresses"
#             )
#             raise Exception("Either draft does not have address tables or loan doesn't have addresses")

#         if kwargs["is_linked_loan"]:
#             loan_ids = ", ".join(loan_ids)
#         else:
#             # For normal loan, we get list of one loan_id so we need to extract this
#             loan_ids = str(loan_ids[0])

#         kwargs["loan_ids"] = loan_ids
#         changed_dt["loan_ids"] = loan_ids

#         notice_data = {
#             "changed_dt": changed_dt,
#             "draft_id": draft_id,
#             "batch_number": batch_number,
#             "document_type": document_type,
#             "notice_type": kwargs.get("notice_type", ""),
#             "case_type": kwargs.get("case_type", ""),
#             "case_id": case_id,
#             "iteration": iteration,
#             "stage_code": stage_code,
#             "is_in_case": kwargs.get("is_in_case", ""),
#             "master_table_dict": master_table_dict,
#             "duplicated_tables_dict": duplicated_tables_dict,
#             "linked_loan_id": linked_loan_id,
#             "languages": languages,
#             "address_set": address_set,
#             "address_tables": address_tables,
#         }
#         return notice_data
#     except Exception as e:
#         raise Exception(f"get_notice_generation_data.exception :: {str(e)}")


# def if_duplicated_table(table_wise_keys, source_key, dup_key):
#     logger.info(f"notice.if_duplicated_table")
#     source_key_identifier = source_key.split("_")[0] + "_" + source_key.split("_")[1]
#     dup_key_identifier = dup_key.split("_")[0] + "_" + dup_key.split("_")[1]
#     return True if table_wise_keys[source_key_identifier] == table_wise_keys[dup_key_identifier] else False


# def duplicate_table_populator(master_table, draft_fields):
#     logger.info(f"notice_service.duplicate_table_populator")
#     table_wise_keys = {}
#     for key, val in master_table.items():
#         table_identifier = key.split("_", 2)[0] + "_" + key.split("_", 2)[1]
#         for j in range(len(draft_fields)):
#             if not table_wise_keys.get(table_identifier):
#                 table_wise_keys[table_identifier] = []
#             if table_identifier in draft_fields[j]:
#                 merge_field = draft_fields[j]
#                 merge_field = (
#                     merge_field.split("_", 4)[-1] if "cell_1" in merge_field else merge_field.split("_", 2)[-1]
#                 )
#                 table_wise_keys[table_identifier].append(merge_field)

#     for key, val in table_wise_keys.items():
#         table_wise_keys[key].sort()
#     keys = {}
#     for key, val in master_table.items():
#         table_name = key.split("_", 4)[-1]
#         if keys.get(table_name):
#             if if_duplicated_table(table_wise_keys, keys[table_name][-2], key):
#                 keys[table_name].insert(-1, key)
#                 keys[table_name][-1] += 1
#         else:
#             keys[table_name] = [key, 1]

#     new_master_table = {}
#     for key, val in keys.items():
#         if val[-1] > 1:
#             source_table = master_table.get(val[0])
#             source_key = val[0]
#             source_table_identifier = source_key.split("_", 2)[0]
#             dups_table = val[1:-1]
#             for i in range(len(dups_table)):
#                 new_master_table[dups_table[i]] = copy.deepcopy(source_table)
#                 del master_table[dups_table[i]]
#                 dup_table_identifier = dups_table[i].split("_", 2)[0]
#                 for index in range(len(new_master_table[dups_table[i]])):
#                     rows = new_master_table[dups_table[i]][index]
#                     for field, value in rows.items():
#                         if source_table_identifier in field:
#                             new_field = field.split(source_table_identifier + "_")[-1]
#                             new_field = dup_table_identifier + "_" + new_field
#                             new_master_table[dups_table[i]][index][new_field] = new_master_table[dups_table[i]][index][
#                                 field
#                             ]

#     return [master_table, new_master_table]


# def get_para_table_dicts(address_set, address_tables, **kwargs):
#     logger.info(f"notice.get_para_table_dicts")
#     logger.debug(f"notice.get_para_table_dicts: {kwargs}")
#     try:
#         company_id = kwargs.get("company_id", "")
#         allocation_month = kwargs.get("allocation_month", "")
#         question_dict = kwargs.get("question_dict", "")
#         notice_type = kwargs.get("notice_type", "")
#         local_batch_directory_path = kwargs["local_batch_directory_path"]
#         local_draft_file_path = os.path.join(local_batch_directory_path, f'{kwargs.get("draft_id","")}.docx')
#         kwargs["local_draft_file_path"] = local_draft_file_path

#         logger.info(f"notice.get_para_table_dicts")
#         logger.debug(f"notice.get_para_table_dicts: {kwargs}")

#         # Change this variable name
#         dt = question_dict

#         changed_dt = {}
#         changed_dt = copy.deepcopy(dt)
#         applicant_count = 1

#         for i in range(applicant_count):
#             changed_dt = {}
#             changed_dt = copy.deepcopy(dt)
#             changed_dt["company_id"] = company_id
#             perform_mailmerge_res = perform_mailmerge(dt, changed_dt, **kwargs)
#             if isinstance(perform_mailmerge_res, str):
#                 return perform_mailmerge_res

#             # if not dt.get("applicant_name", ""):
#             #     raise Exception("get_para_table_dicts.exception: Missing applicant_name")

#             changed_dt["name"] = dt.get("applicant_name", "")
#             changed_dt["contact_number"] = str(dt.get("applicant_contact_number", ""))

#             if dt.get("applicant_address", "") not in [None, "", []]:
#                 changed_dt["address_text"] = dt["applicant_address"][i].get("applicant_address_text", "")
#                 changed_dt["city"] = dt["applicant_address"][i].get("applicant_city", "")
#                 changed_dt["state"] = dt["applicant_address"][i].get("applicant_state", "")
#                 changed_dt["pincode"] = str(dt["applicant_address"][i].get("applicant_pincode", ""))
#             changed_dt["address_index"] = i
#             changed_dt["applicant_type"] = applicant_type = "applicant"
#             changed_dt["allocation_month"] = allocation_month

#             if dt.get("notice_reference_number", "") in [None, "", []]:
#                 get_notice_reference_number(dt, changed_dt, "notice_reference_number", notice_type)

#             if dt.get("applicant_notice_reference_number", "") in [None, "", []]:
#                 get_notice_reference_number(dt, changed_dt, "applicant_notice_reference_number", notice_type)

#             if dt.get("co_applicant_notice_reference_number", "") in [None, "", []]:
#                 get_notice_reference_number(dt, changed_dt, "co_applicant_notice_reference_number", notice_type)

#             kwargs["company_id"] = company_id
#             kwargs["changed_dt"] = changed_dt
#             table_dict = populate_table_rows(address_set, address_tables, **kwargs)
#         return [changed_dt, table_dict]
#     except Exception as e:
#         raise Exception(f"{str(e)}")


# def populate_table_rows(address_set, address_tables, **kwargs):
#     logger.info("notice.populate_table_rows")
#     question_dict = kwargs.get("question_dict", "")
#     # doc = MailMerge(kwargs['local_draft_file_path'])
#     # fields = list(doc.get_merge_fields())
#     fields = kwargs.get("draft_fields", [])
#     # doc.close()
#     cell_1s = []
#     row_dict = {}
#     fields.sort()
#     table_no = 0
#     table_map = {}
#     parsed_tables = set()
#     table_identifier = ""

#     for i in range(len(fields)):
#         if "table" in fields[i]:
#             row_dict[fields[i]] = ""
#             if "cell_1" in fields[i]:
#                 cell_1s.append((fields[i]))

#     kwargs["cell_1s"] = cell_1s

#     for i in range(len(cell_1s)):
#         table_identifier = cell_1s[i].split("_", 2)[0] + "_" + cell_1s[i].split("_", 2)[1]
#         tag = cell_1s[i].split("_", 4)[-1]
#         tag = tag.split("_")[0]
#         if tag == "co":
#             tag = "co_applicant"
#         for j in range(len(fields)):
#             if table_identifier in fields[j] and ("address_text" in fields[j] or "builder_address" in fields[j]):
#                 address_tables.add(tag)

#         if tag in address_tables:
#             if not table_map.get(tag):
#                 table_map[tag] = []
#             table_map[tag].append(str(table_no + 1))
#         else:
#             if not table_map.get(cell_1s[i]):
#                 table_map[cell_1s[i]] = []
#             table_map[cell_1s[i]].append(str(table_no + 1))
#         table_no += 1

#     table_dict = {}
#     kwargs["row_dict"] = row_dict
#     kwargs["fields"] = fields
#     kwargs["table_map"] = table_map
#     if question_dict.get("property", "") not in ("", [], None):
#         kwargs["property_address"] = question_dict.get("property", "")

#     for i in range(len(cell_1s)):
#         merge_list = []
#         table_name = cell_1s[i].split("_", 4)[-1].split("_")[0]
#         kwargs["current_cell_1_cell"] = cell_1s[i]
#         if table_name == "co":
#             table_name += "_applicant"
#         kwargs["table_name"] = table_name
#         if table_name == "applicant":
#             merge_list = mediator(question_dict, address_set, **kwargs)
#         elif table_name in address_tables:
#             qd = question_dict.get(table_name, "")
#             if qd not in ("", None, []):
#                 merge_list = mediator(qd, address_set, **kwargs)
#         else:
#             merge_list = populate_from_changed_dt(**kwargs)

#         table_dict[cell_1s[i]] = merge_list
#         del merge_list

#     return table_dict


# def populate_from_changed_dt(**kwargs):
#     fields = kwargs.get("fields", "")
#     row_dict = kwargs.get("row_dict", "")
#     merge_list = []
#     changed_dt = kwargs.get("changed_dt", "")
#     merge_dict_copy = row_dict.copy()
#     for i in range(len(fields)):
#         if "cell_1" in fields[i]:
#             merge_field = fields[i].split("_", 4)[-1]
#         else:
#             merge_field = fields[i].split("_", 2)[-1]

#         merge_dict_copy[fields[i]] = changed_dt.get(merge_field, "")

#     merge_list.append(merge_dict_copy)
#     del merge_dict_copy

#     return merge_list


# def mediator(qd, address_set, **kwargs):
#     logger.info("notice.mediator")

#     if isinstance(qd, list):
#         table_data = []
#         for l in range(len(qd)):
#             response = []
#             qd1 = qd[l]
#             response = populate_rows(qd1, address_set, **kwargs)
#             for k in range(len(response)):
#                 table_data.append(response[k])
#             del response

#         return table_data
#     else:
#         return populate_rows(qd, address_set, **kwargs)


# def populate_rows(qd, address_set, **kwargs):
#     logger.info("notice.populate_rows")

#     merge_list = []
#     table_name = kwargs["table_name"]
#     fields = kwargs["fields"]
#     row_dict = kwargs["row_dict"]
#     company_id = kwargs["company_id"]
#     table_map = kwargs["table_map"]
#     property_address = {}
#     loan_address_map = kwargs["loan_address_map"]
#     loan_id = kwargs["loan_id"]
#     primary_name, primary_address, primary_address_type = "", "", ""
#     if kwargs.get("property_address", "") not in ("", [], None):
#         property_address = kwargs.get("property_address", "")
#     cell_1s = kwargs["cell_1s"]

#     address_name = ""
#     KEY = table_name + "_address"
#     borrower_type = table_name.lower()
#     if qd.get(KEY, ""):
#         for z in range(len(qd.get(KEY, ""))):
#             merge_dict_copy = row_dict.copy()
#             if qd[KEY][z].get(table_name + "_address_text", "") not in ["", None, []]:
#                 address_name = qd[KEY][z].get(table_name + "_address_text", "") + qd.get(table_name + "_name", "")
#                 if address_name not in address_set:
#                     if qd.get(table_name + "_name", "") or table_name in [
#                         "property",
#                         "auctioned",
#                     ]:
#                         address_set.add(address_name)

#                     for i in range(len(fields)):
#                         if "cell_1" in fields[i]:
#                             tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                         else:
#                             tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                         if isinstance(qd.get(tag, ""), str) and qd.get(tag, "") not in [
#                             "",
#                             None,
#                         ]:
#                             val = remove_special_characters(qd.get(tag, ""))
#                         else:
#                             val = qd.get(tag, "")
#                         merge_dict_copy[fields[i]] = val

#                         if merge_dict_copy[fields[i]] == "":
#                             if isinstance(qd[KEY][z].get(tag, ""), str) and qd[KEY][z].get(tag, "") not in ["", None]:
#                                 val = remove_special_characters(qd[KEY][z].get(tag, ""))
#                             else:
#                                 val = qd[KEY][z].get(tag, "")
#                             merge_dict_copy[fields[i]] = val

#                         # loan_address_map_population
#                         if "name" in tag and val != "":
#                             primary_name = val
#                         if "address_text" in tag and val != "":
#                             primary_address = val

#                         address_type_key = table_name + "_address_type"

#                         if qd[KEY][z].get(address_type_key, ""):
#                             primary_address_type = qd[KEY][z][address_type_key]

#                         if tag == "borrower_type" and fields[i].split("_", 2)[1] in table_map[table_name]:
#                             merge_dict_copy[fields[i]] = "(" + table_name.title() + ")"

#                         kwargs["field"] = fields[i]
#                         merge_dict_copy["loan_id"] = loan_id
#                         # merge_dict_copy = iifl_check(qd, merge_dict_copy, **kwargs)

#                         if z != 0 and any(value in fields[i] for value in ["contact_number", "email"]):
#                             merge_dict_copy[fields[i]] = ""

#                         if table_name == "applicant" and "reference" in tag:
#                             changed_dt_key = tag.split("_", 2)[-1] + "_" + str(z + 1)
#                             merge_dict_copy[fields[i]] = kwargs["changed_dt"].get(changed_dt_key, "")

#                     merge_list.append(merge_dict_copy)
#                     del merge_dict_copy
#                 else:
#                     for i in range(len(fields)):
#                         if "cell_1" in fields[i]:
#                             tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                         else:
#                             tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                         if isinstance(qd.get(tag, ""), str) and qd.get(tag, "") not in [
#                             "",
#                             None,
#                         ]:
#                             val = remove_special_characters(qd.get(tag, ""))
#                         else:
#                             val = qd.get(tag, "")
#                         merge_dict_copy[fields[i]] = val
#                         if merge_dict_copy[fields[i]] == "":
#                             if isinstance(qd[KEY][z].get(tag, ""), str) and qd[KEY][z].get(tag, "") not in ["", None]:
#                                 val = remove_special_characters(qd[KEY][z].get(tag, ""))
#                             else:
#                                 val = qd[KEY][z].get(tag, "")
#                             merge_dict_copy[fields[i]] = val

#                         # loan_address_map_population
#                         if "name" in tag and val != "":
#                             primary_name = val
#                         if "address_text" in tag and val != "":
#                             primary_address = val

#                         address_type_key = table_name + "_address_type"

#                         if qd[KEY][z].get(address_type_key, ""):
#                             primary_address_type = qd[KEY][z][address_type_key]

#                         if tag == "borrower_type" and fields[i].split("_", 2)[1] in table_map[table_name]:
#                             merge_dict_copy[fields[i]] = "(" + table_name.title() + ")"

#                         if z != 0 and any(value in fields[i] for value in ["contact_number", "email"]):
#                             merge_dict_copy[fields[i]] = ""

#             primary_key = (primary_name + primary_address, borrower_type, primary_address_type)
#             loan_id = kwargs["loan_id"]
#             if not loan_address_map.get(loan_id):
#                 loan_address_map[loan_id] = [primary_key]
#             else:
#                 if primary_key not in loan_address_map[loan_id]:
#                     loan_address_map[loan_id].append(primary_key)

#     elif (
#         qd.get(table_name + "_builder_address") not in ("", [], None)
#         and f"{table_name}_builder" in kwargs["current_cell_1_cell"]
#     ):
#         address_name = qd.get(table_name + "_builder_name", "") + qd[table_name + "_builder_address"]
#         if address_name not in address_set:
#             address_set.add(address_name)
#             merge_dict_copy = row_dict.copy()
#             for i in range(len(fields)):
#                 if "cell_1" in fields[i]:
#                     tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                 else:
#                     tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                 if isinstance(qd.get(tag, ""), str) and qd.get(tag, "") not in [
#                     "",
#                     None,
#                 ]:
#                     val = remove_special_characters(qd.get(tag, ""))
#                 else:
#                     val = qd.get(tag, "")

#                 merge_dict_copy[fields[i]] = val
#             merge_list.append(merge_dict_copy)
#             del merge_dict_copy

#     elif qd.get(table_name + "_address_text", "") not in ("", [], None):
#         address_name = qd.get(table_name + "_address_text", "") + qd.get(table_name + "_name", "")
#         merge_dict_copy = row_dict.copy()

#         if address_name not in address_set:
#             if qd.get(table_name + "_name", "") or table_name in [
#                 "property",
#                 "auctioned",
#             ]:
#                 address_set.add(address_name)
#             for i in range(len(fields)):
#                 if "cell_1" in fields[i]:
#                     tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                 else:
#                     tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                 if isinstance(qd.get(tag, ""), str) and qd.get(tag, "") not in [
#                     "",
#                     None,
#                 ]:
#                     val = remove_special_characters(qd.get(tag, ""))
#                 else:
#                     val = qd.get(tag, "")
#                 merge_dict_copy[fields[i]] = val

#                 # loan_address_map_population
#                 if "name" in tag and val != "":
#                     primary_name = val
#                 if "address_text" in tag and val != "":
#                     primary_address = val

#                 address_type_key = table_name + "_address_type"
#                 if qd.get(address_type_key, ""):
#                     primary_address_type = qd[address_type_key]

#                 if tag == "borrower_type" and fields[i].split("_", 2)[1] in table_map[table_name]:
#                     merge_dict_copy[fields[i]] = "(" + table_name.title() + ")"

#                 kwargs["field"] = fields[i]
#                 merge_dict_copy["loan_id"] = loan_id
#                 # merge_dict_copy = iifl_check(qd, merge_dict_copy, **kwargs)

#                 if table_name == "applicant" and "reference" in tag:
#                     changed_dt_key = tag.split("_", 2)[-1] + "_" + str(z + 1)
#                     merge_dict_copy[fields[i]] = kwargs["changed_dt"].get(changed_dt_key, "")

#             merge_list.append(merge_dict_copy)
#             del merge_dict_copy
#         else:
#             for i in range(len(fields)):
#                 if "cell_1" in fields[i]:
#                     tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                 else:
#                     tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                 if isinstance(qd.get(tag, ""), str) and qd.get(tag, "") not in [
#                     "",
#                     None,
#                 ]:
#                     val = remove_special_characters(qd.get(tag, ""))
#                 else:
#                     val = qd.get(tag, "")
#                 merge_dict_copy[fields[i]] = val

#                 # loan_address_map_population
#                 if "name" in tag and val != "":
#                     primary_name = val
#                 if "address_text" in tag and val != "":
#                     primary_address = val

#                 address_type_key = table_name + "_address_type"
#                 if qd.get(address_type_key, ""):
#                     primary_address_type = qd[address_type_key]

#                 if tag == "borrower_type" and fields[i].split("_", 2)[1] in table_map[table_name]:
#                     merge_dict_copy[fields[i]] = "(" + table_name.title() + ")"

#         primary_key = (primary_name + primary_address, borrower_type, primary_address_type)
#         loan_id = kwargs["loan_id"]
#         if not loan_address_map.get(loan_id):
#             loan_address_map[loan_id] = [primary_key]
#         else:
#             if primary_key not in loan_address_map[loan_id]:
#                 loan_address_map[loan_id].append(primary_key)
#     kwargs["loan_address_map"] = loan_address_map

#     if company_id in (NOTICE_COMPANIES_WITH_PROPERTY_ADDRESS_HANDLING):
#         if (
#             len(property_address) > 0
#             and table_name != "property"
#             and (
#                 "property_address_text" in [merge_field.split("_", 4)[-1] for merge_field in cell_1s]
#                 or any([merge_field.startswith("property_address_text") for merge_field in kwargs["draft_fields"]])
#             )
#         ):
#             populate_property_address(qd, merge_list, address_set, **kwargs)

#     return merge_list


# def perform_mailmerge(dt, changed_dt, **kwargs):
#     logger.info("notice.perform_mailmerge")
#     for key in dt:
#         if dt[key] in (None, [], "nan", "Nan", "none", "None", "NAN", "NONE"):
#             changed_dt[key] = ""
#         elif isinstance(dt[key], int) or isinstance(dt[key], float):
#             # Logic of open_int_or_float
#             changed_dt[key] = open_int_or_float(key, dt[key])
#         elif isinstance(dt[key], str) and dt[key] not in ["", None]:
#             # Remove special characters for all the values except email
#             if any(
#                 key_str in key
#                 for key_str in [
#                     "chassis_number",
#                     "engine_number",
#                     "vehicle_registration_number",
#                     "make_and_model",
#                 ]
#             ):
#                 dt[key] = remove_special_characters(dt[key], "*")
#             elif all([False if str in key else True for str in ["email", "payment_link"]]):
#                 dt[key] = remove_special_characters(dt[key])
#             if key == "barcode":
#                 dt[key] = "(" + dt[key].upper() + ")"
#             changed_dt[key] = open_str(key, dt[key])

#         if dt[key] not in (None, "", []):
#             if "date" in key:
#                 changed_dt[key] = open_date(key, dt[key])

#             if isinstance(dt[key], list) and key != "defaults":
#                 open_list(changed_dt, dt[key], -1, -1)
#                 del changed_dt[key]
#             elif key == "defaults":
#                 del changed_dt["defaults"]
#             elif isinstance(dt[key], dict):
#                 open_dict(changed_dt, dt[key], 0, 0)

#     if "total_claim_amount" in dt and dt["total_claim_amount"] not in ["", None, []]:
#         changed_dt["total_claim_amount"] = format(dt["total_claim_amount"], ".2f")

#     changed_dt["present_date"] = datetime.now().strftime("%d/%m/%Y")

#     merge_fields = kwargs["draft_fields"]

#     in_words = [i for i in merge_fields if "_in_words" in i]

#     for var in in_words:
#         if "cell_1" in var:
#             var = var.split("_", 4)[-1]
#         elif "table_" in var:
#             var = var.split("_", 2)[-1]
#         amount_key = var.split("_in_words")[0]
#         if amount_key in changed_dt and changed_dt[amount_key] not in ("", None, []):
#             changed_dt[var] = convert_num2words(changed_dt[amount_key])
#         elif (
#             "defaults" in dt
#             and amount_key in dt["defaults"][-1]
#             and dt["defaults"][-1][amount_key] not in ("", None, [])
#         ):
#             changed_dt[var] = convert_num2words(dt["defaults"][-1][amount_key])

#     changed_dt2 = copy.deepcopy(changed_dt)
#     for key, val in changed_dt2.items():
#         if key.endswith("_1_1"):
#             tag = key.split("_1_1")[0] + "_1"
#             changed_dt[tag] = val
#         elif key.endswith("_1") and not key.startswith("applicant"):
#             tag = key.split("_1")[0]
#             changed_dt[tag] = val

#     if changed_dt.get("company_id", "") == "d2d50f3e-4b79-40d7-b9dd-6cadb0b30141":
#         axis_check(changed_dt)

#     return True


# def remove_special_characters(input_str, additional_exceptions=""):
#     logger.info("notice.remove_special_characters")
#     logger.debug(f"input_str : {input_str}")
#     input_str = input_str.replace("\\n", "")
#     input_str = input_str.replace("\n", "")
#     input_str = input_str.replace("/n", "")
#     # Matches individual characters of the input_str, the pattern is that except A-Z, a-z, 0-9 replace all other listed characters with ''(empty character)
#     exception_str = f"[^A-Za-z0-9+ _,-/.@:()\[\]&#{additional_exceptions}]+"
#     output_str = re.sub(exception_str, "", input_str)
#     logger.debug(f"output_str : {output_str}")
#     return output_str


# def get_notice_reference_number(dt, changed_dt, key, notice_type):

#     if dt.get("notice_date", "") not in ["", None, []] and is_date(dt.get("notice_date", "")):
#         try:
#             changed_dt[key] = (
#                 str(dt["loan_id"])
#                 + "-"
#                 + notice_type.upper()
#                 + "-"
#                 + datetime.strptime(dt["notice_date"], "%Y-%m-%d").strftime("%d%m%Y")
#                 + "-"
#                 + str(random.randint(1000, 9999))
#             )
#         except Exception as e:
#             logger.error(f"date type is wrong. error handled: {str(e)}")
#             changed_dt[key] = (
#                 str(dt["loan_id"])
#                 + "-"
#                 + notice_type.upper()
#                 + "-"
#                 + datetime.now().strftime("%d%m%Y")
#                 + "-"
#                 + str(random.randint(1000, 9999))
#             )

#     else:
#         changed_dt[key] = (
#             str(dt["loan_id"])
#             + "-"
#             + notice_type.upper()
#             + "-"
#             + datetime.now().strftime("%d%m%Y")
#             + "-"
#             + str(random.randint(1000, 9999))
#         )


# def populate_property_address(qd, merge_list, address_set, **kwargs):
#     table_name = kwargs["table_name"]
#     fields = kwargs["fields"]
#     row_dict = kwargs["row_dict"]
#     company_id = kwargs["company_id"]
#     table_map = kwargs["table_map"]
#     property_address = {}
#     loan_address_map = kwargs["loan_address_map"]
#     loan_id = kwargs["loan_id"]
#     address_name = ""
#     if kwargs.get("property_address", "") not in ("", [], None):
#         property_address = kwargs.get("property_address", "")

#     for p_address in property_address:
#         if p_address.get("property_address_text", "") not in ("", [], None):
#             merge_dict_copy = row_dict.copy()
#             for i in range(len(fields)):
#                 if "cell_1" in fields[i]:
#                     tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                 else:
#                     tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                 if "name" in tag and fields[i].split("_", 2)[1] in table_map[table_name]:
#                     address_name = p_address.get("property_address_text", "") + qd.get(tag, "")
#                     primary_name = qd.get(tag, "")
#                     break

#             if address_name not in address_set:
#                 address_set.add(address_name)
#                 for i in range(len(fields)):
#                     if "cell_1" in fields[i]:
#                         tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                     else:
#                         tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                     if "name" in tag:
#                         merge_dict_copy[fields[i]] = qd.get(tag, "")
#                     elif "borrower_type" not in tag:
#                         tag = tag.split(table_name + "_")[-1]
#                         tag = "property_" + tag
#                         merge_dict_copy[fields[i]] = p_address.get(tag, "")
#                         if "address_text" in tag and p_address.get(tag, "") != "":
#                             primary_address = p_address.get(tag, "")

#                     if tag == "borrower_type" and fields[i].split("_", 2)[1] in table_map[table_name]:
#                         merge_dict_copy[fields[i]] = "(" + table_name.title() + ")"

#                     kwargs["field"] = fields[i]

#                 merge_list.append(merge_dict_copy)
#                 del merge_dict_copy

#             else:
#                 for i in range(len(fields)):
#                     if "cell_1" in fields[i]:
#                         tag = fields[i].split("_", 4)[-1].strip("_").strip(" ")
#                     else:
#                         tag = fields[i].split("_", 2)[-1].strip("_").strip(" ")

#                     if "name" in tag:
#                         merge_dict_copy[fields[i]] = qd.get(tag, "")
#                     elif "borrower_type" not in tag:
#                         tag = tag.split(table_name + "_")[-1]
#                         tag = "property_" + tag
#                         merge_dict_copy[fields[i]] = p_address.get(tag, "")
#                         if "address_text" in tag and p_address.get(tag, "") != "":
#                             primary_address = p_address.get(tag, "")

#             primary_key = primary_name + primary_address
#             loan_id = kwargs["loan_id"]
#             borrower_type = table_name

#             if not loan_address_map.get(loan_id):
#                 loan_address_map[loan_id] = [(primary_key, borrower_type, primary_address)]
#             else:
#                 loan_address_map[loan_id].append((primary_key, borrower_type, primary_address))
#             kwargs["loan_address_map"] = loan_address_map


# def open_int_or_float(key, value):
#     if (
#         "amount" in key
#         or "charges" in key
#         or "dues" in key
#         or key in ["expected_emi", "late_fee", "other_penalty", "overdue_charges"]
#     ) and value not in ("", None, []):
#         res = format(value, ".2f")
#     else:
#         res = str(value)

#     return res


# def open_str(key, value):
#     if (
#         "amount" in key
#         or "charges" in key
#         or "dues" in key
#         or key in ["expected_emi", "late_fee", "other_penalty", "overdue_charges"]
#     ) and value not in ("", None, []):
#         try:
#             res = format(float(value), ".2f")
#         except:
#             res = str(value.strip())
#     else:
#         res = str(value.strip())

#     return res


# def open_date(key, value):
#     changed_value = ""
#     logger.debug(f"open_date.value:: {value}")
#     if isinstance(value, datetime):
#         try:
#             changed_value = value.strftime("%d/%m/%Y")
#         except Exception as e:
#             logger.error(f"datetime type date conversion error:: {str(e)}")
#             changed_value = str(value)
#     elif isinstance(value, str):
#         try:
#             changed_value = datetime.strptime(value.split(" ")[0], "%Y-%m-%d").strftime("%d/%m/%Y")
#         except Exception as e:
#             logger.error(f"str type date conversion error:: {str(e)}")
#             changed_value = str(value.split(" ")[0])
#     elif value in ["", None]:
#         return ""

#     return changed_value


# def open_dict(changed_dt, dicti, list_no, i):
#     for key, val in dicti.items():
#         if isinstance(val, list):
#             open_list(changed_dt, val, i, i)
#         elif isinstance(val, dict):
#             open_dict(changed_dt, val, list_no, i)
#         elif isinstance(val, int) or isinstance(val, float):
#             if list_no > -1:
#                 changed_dt[key + "_" + str(list_no + 1) + "_" + str(i + 1)] = open_int_or_float(key, val)
#             else:
#                 changed_dt[key + "_" + str(i + 1)] = open_int_or_float(key, val)
#         elif "date" in key and val not in ("", None, []):
#             if list_no > -1:
#                 changed_dt[key + "_" + str(list_no + 1) + "_" + str(i + 1)] = open_date(key, val)
#             else:
#                 changed_dt[key + "_" + str(i + 1)] = open_date(key, val)
#         elif isinstance(val, str) and val not in ["", None]:
#             if "email" not in key:
#                 val = remove_special_characters(val)
#             if list_no > -1:
#                 changed_dt[key + "_" + str(list_no + 1) + "_" + str(i + 1)] = open_str(key, val)
#             else:
#                 changed_dt[key + "_" + str(i + 1)] = open_str(key, val)
#         elif val in (None, []):
#             if list_no > -1:
#                 changed_dt[key + "_" + str(list_no + 1) + "_" + str(i + 1)] = ""
#             else:
#                 changed_dt[key + "_" + str(i + 1)] = ""
#         else:
#             if list_no > -1:
#                 changed_dt[key + "_" + str(list_no + 1) + "_" + str(i + 1)] = str(val.strip())
#             else:
#                 changed_dt[key + "_" + str(i + 1)] = str(val.strip())

#         if "applicant_address_text" in key:
#             co = ""
#             if "co" in key:
#                 co = "co_"

#             val = format_address(
#                 dicti.get(co + "applicant_address_text", ""),
#                 dicti.get(co + "applicant_state", ""),
#                 str(dicti.get(co + "applicant_pincode", "")),
#                 dicti.get(co + "applicant_city", ""),
#             )
#             if list_no > -1:
#                 changed_dt[key + "_" + str(list_no + 1) + "_" + str(i + 1)] = val
#             else:
#                 changed_dt[key + "_" + str(i + 1)] = val


# def open_list(changed_dt, lst, list_no, iterator):
#     for i in range(len(lst)):
#         if isinstance(lst[i], dict):
#             open_dict(changed_dt, lst[i], list_no, i)
#         elif isinstance(lst[i], list):
#             open_list(changed_dt, lst[i], list_no, i)


# def convert_num2words(num):
#     logger.info("notice.convert_num2words")
#     logger.debug(f"notice.convert_num2words.num: '{num}'")

#     try:
#         decimal_val = ""
#         integer_val = str(num).split(".")[0]

#         if isinstance(num, str):
#             try:
#                 num = float(num)
#             except Exception as e:
#                 raise Exception(f"notice.convert_num2words.string_amount_conversion.exception:: {str(e)}")

#         if isinstance(num, float):
#             num = str(num) + "0"
#             integer_val, decimal_val = num.split(".")
#             decimal_val = decimal_val[:2]
#             num = integer_val + "." + "".join(decimal_val)
#             paise2words = num2words(decimal_val, lang="en_IN")

#         rupees2words = num2words(integer_val, lang="en_IN")

#         num2words_string = "Rupees " + rupees2words

#         if decimal_val and decimal_val != "00":
#             num2words_string = (
#                 num2words_string + " and " + paise2words + (" paisa" if decimal_val[0] == "0" else " paise")
#             )
#         formatted_num2words_string = " ".join(
#             " ".join(str.strip().split("-")) for str in num2words_string.title().split(",")
#         )
#         formatted_words_list = formatted_num2words_string.split("And")

#         if len(formatted_words_list) > 1:
#             formatted_num2words_string = (
#                 formatted_words_list[0].strip()
#                 + " "
#                 + formatted_words_list[1].strip()
#                 + " and "
#                 + formatted_words_list[2].strip()
#                 if len(formatted_words_list) >= 3
#                 else formatted_words_list[0].strip() + " and " + formatted_words_list[1].strip()
#             )
#         logger.debug(f"notice.convert_num2words.formatted_num2words_string: '{formatted_num2words_string}'")
#         return formatted_num2words_string + " Only"
#     except:
#         logger.error(f"error: unsupported format of variable 'num', other than int, float and string e.g.:(6,6.0,'6') ")
#         return ""


# def is_date(string, fuzzy=False):
#     try:
#         parse(string, fuzzy=fuzzy)
#         return True
#     except ValueError:
#         return False


# def axis_check(changed_dt):
#     logger.info("notice.axis_check")
#     logger.debug(
#         f"applicant_address_text_1: {changed_dt.get('applicant_address_text_1','')}, applicant_address_text_2: {changed_dt.get('applicant_address_text_1','')}, 'applicant_address_text_3': {changed_dt.get('applicant_address_text_3','')}"
#     )
#     permanent_address_text, current_adddres_text, work_address_text = "", "", ""
#     permanent_address_city, current_address_city, work_address_city = "", "", ""
#     permanent_address_state, current_address_state, work_address_state = "", "", ""
#     permanent_address_pincode, current_address_pincode, work_address_pincode = (
#         "",
#         "",
#         "",
#     )
#     (
#         permanent_address_text_index,
#         current_address_text_index,
#         work_address_text_index,
#     ) = ("", "", "")
#     for i in range(9):
#         if changed_dt.get("applicant_address_type_" + str(i + 1), "").lower() == "permanent":
#             permanent_address_text = changed_dt.get("applicant_address_text_" + str(i + 1), "")
#             permanent_address_city = changed_dt.get("applicant_city_" + str(i + 1), "")
#             permanent_address_state = changed_dt.get("applicant_state" + str(i + 1), "")
#             permanent_address_pincode = changed_dt.get("applicant_pincode" + str(i + 1), "")
#             permanent_address_text_index = str(i + 1)
#         elif changed_dt.get("applicant_address_type_" + str(i + 1), "").lower() == "current":
#             current_adddres_text = changed_dt.get("applicant_address_text_" + str(i + 1), "")
#             current_address_city = changed_dt.get("applicant_city_" + str(i + 1), "")
#             current_address_state = changed_dt.get("applicant_state" + str(i + 1), "")
#             current_address_pincode = changed_dt.get("applicant_pincode" + str(i + 1), "")
#             current_address_text_index = str(i + 1)
#         elif changed_dt.get("applicant_address_type_" + str(i + 1), "").lower() == "work":
#             work_adddres_text = changed_dt.get("applicant_address_text_" + str(i + 1), "")
#             work_address_city = changed_dt.get("applicant_city_" + str(i + 1), "")
#             work_address_state = changed_dt.get("applicant_state" + str(i + 1), "")
#             work_address_pincode = changed_dt.get("applicant_pincode" + str(i + 1), "")
#             work_address_text_index = str(i + 1)
#     if permanent_address_text:
#         (
#             changed_dt["applicant_address_text_1"],
#             changed_dt["applicant_address_text_" + permanent_address_text_index],
#         ) = (
#             changed_dt["applicant_address_text_" + permanent_address_text_index],
#             changed_dt["applicant_address_text_1"],
#         )
#         (changed_dt["applicant_city_1"], changed_dt["applicant_city_" + permanent_address_text_index],) = (
#             changed_dt["applicant_city_" + permanent_address_text_index],
#             changed_dt["applicant_city_1"],
#         )
#         (changed_dt["applicant_state_1"], changed_dt["applicant_state_" + permanent_address_text_index],) = (
#             changed_dt["applicant_state_" + permanent_address_text_index],
#             changed_dt["applicant_state_1"],
#         )
#         (changed_dt["applicant_pincode_1"], changed_dt["applicant_pincode_" + permanent_address_text_index],) = (
#             changed_dt["applicant_pincode_" + permanent_address_text_index],
#             changed_dt["applicant_pincode_1"],
#         )
#         (
#             changed_dt["applicant_address_type_1"],
#             changed_dt["applicant_address_type_" + permanent_address_text_index],
#         ) = (
#             changed_dt["applicant_address_type_" + permanent_address_text_index],
#             changed_dt["applicant_address_type_1"],
#         )
#     elif current_adddres_text:
#         (
#             changed_dt["applicant_address_text_1"],
#             changed_dt["applicant_address_text_" + current_address_text_index],
#         ) = (
#             changed_dt["applicant_address_text_" + current_address_text_index],
#             changed_dt["applicant_address_text_1"],
#         )
#         (changed_dt["applicant_city_1"], changed_dt["applicant_city_" + current_address_text_index],) = (
#             changed_dt["applicant_city_" + current_address_text_index],
#             changed_dt["applicant_city_1"],
#         )
#         (changed_dt["applicant_state_1"], changed_dt["applicant_state_" + current_address_text_index],) = (
#             changed_dt["applicant_state_" + current_address_text_index],
#             changed_dt["applicant_state_1"],
#         )
#         (changed_dt["applicant_pincode_1"], changed_dt["applicant_pincode_" + current_address_text_index],) = (
#             changed_dt["applicant_pincode_" + current_address_text_index],
#             changed_dt["applicant_pincode_1"],
#         )
#         (
#             changed_dt["applicant_address_type_1"],
#             changed_dt["applicant_address_type_" + current_address_text_index],
#         ) = (
#             changed_dt["applicant_address_type_" + current_address_text_index],
#             changed_dt["applicant_address_type_1"],
#         )
#     elif work_address_text:
#         (changed_dt["applicant_address_text_1"], changed_dt["applicant_address_text_" + work_address_text_index],) = (
#             changed_dt["applicant_address_text_" + work_address_text_index],
#             changed_dt["applicant_address_text_1"],
#         )
#         (changed_dt["applicant_city_1"], changed_dt["applicant_city_" + work_address_text_index],) = (
#             changed_dt["applicant_city_" + work_address_text_index],
#             changed_dt["applicant_city_1"],
#         )
#         (changed_dt["applicant_state_1"], changed_dt["applicant_state_" + work_address_text_index],) = (
#             changed_dt["applicant_state_" + work_address_text_index],
#             changed_dt["applicant_state_1"],
#         )
#         (changed_dt["applicant_pincode_1"], changed_dt["applicant_pincode_" + work_address_text_index],) = (
#             changed_dt["applicant_pincode_" + work_address_text_index],
#             changed_dt["applicant_pincode_1"],
#         )
#         (changed_dt["applicant_address_type_1"], changed_dt["applicant_address_type_" + work_address_text_index],) = (
#             changed_dt["applicant_address_type_" + work_address_text_index],
#             changed_dt["applicant_address_type_1"],
#         )


# def format_address(address_text, state, pincode, city):
#     logger.info("notice.format_address")
#     logger.debug(f"address_text: '{address_text}', state: '{state}', pincode: '{pincode}', city: '{city}'")
#     # if state != 'Delhi':
#     #   address_text = address_text.replace(state, '')
#     # if city != 'Delhi':
#     #   address_text = address_text.replace(city, '')

#     if address_text in ("", [], "nan", "Nan", "none", "None", "NAN", "NONE", None):
#         logger.debug(f"address_text: '{address_text}'")
#         address_text = ""
#         return address_text

#     if pincode:
#         address_text = address_text.replace(pincode, "")
#     address_text = address_text.replace("  ", " ")
#     address_text = address_text.replace(" ,", ",")
#     address_text = address_text.replace(",,,", "")
#     address_text = address_text.replace(",,", ",")
#     address_text = address_text.replace(", -", "")
#     address_text = address_text.replace(",", ", ")
#     address_text = address_text.replace("\\n", "")
#     address_text = address_text.replace("\n", "")
#     address_text = address_text.replace("/n", "")
#     address_text = address_text.replace("State", "")
#     address_text = address_text.replace("City", "")
#     address_text = re.sub("[^A-Za-z0-9 ,:\-/.@()\[\]&#+]+", "", address_text)
#     address_text = address_text.replace("  ", " ")
#     address_text = address_text.strip()

#     if address_text in ("", [], "nan", "Nan", "none", "None", "NAN", "NONE", None):
#         logger.debug(f"address_text: '{address_text}'")
#         address_text = ""
#         return address_text

#     if address_text[-1] in (",", " ", "-"):
#         logger.debug(f"address_text: '{address_text}'")
#         return address_text[:-1]
#     if address_text[-2:] == ", ":
#         logger.debug(f"address_text: '{address_text}'")
#         return address_text[:-2]
#     logger.debug(f"address_text: '{address_text}'")
#     return address_text


# def upload_notices_to_s3(**kwargs):
#     logger.info("notice.upload_notices_to_s3")
#     logger.info(f"notice.upload_notices_to_s3.batch_id :: {kwargs['batch_id']}")

#     redis_instance = redis.Redis(host=REDIS["HOST"], port=REDIS["PORT"])
#     files_s3_map = kwargs.get("files_s3_map", {})
#     filenames = kwargs["filenames"]
#     BUCKET_NAME = kwargs["BUCKET_NAME"]
#     S3_BUCKET_ENDPOINT = kwargs["S3_BUCKET_ENDPOINT"]
#     s3_link_ids = []
#     s3_object_urls = []
#     s3_folder = "preview_notice_links" if kwargs.get("preview") else "notice_links"

#     try:
#         for output_file in filenames:
#             file = open(output_file, "rb")
#             s3_link_uuid = kwargs["s3_link_uuid"] if kwargs.get("s3_link_uuid") else gen_short_key(12)
#             s3.upload_file(
#                 file.name,
#                 BUCKET_NAME,
#                 s3_folder + "/" + s3_link_uuid + ".pdf",
#                 ExtraArgs={"ACL": "public-read", "ContentType": "application/pdf"},
#             )

#             s3_object_url = f"{S3_BUCKET_ENDPOINT.rstrip('/')}/{s3_folder}/{s3_link_uuid}.pdf"
#             # s3.upload_file(file.name, BUCKET_NAME, 'notice_links/' + s3_link_uuid + '.pdf', ExtraArgs={'ACL':'public-read'})

#             # s3.copy_object(Bucket = BUCKET_NAME, Key = 'notice_links/' + s3_link_uuid + '.pdf', CopySource = BUCKET_NAME + '/' + 'notice_links/' + s3_link_uuid + '.pdf', ContentType='application/pdf', MetadataDirective='REPLACE')

#             # s3_object_url = f"{S3_BUCKET_ENDPOINT.rstrip('/')}/notice_links/{s3_link_uuid}.pdf"
#             # s3.put_object_acl(ACL='public-read', Bucket= BUCKET_NAME, Key='notice_links/' + s3_link_uuid + '.pdf')
#             output_file = output_file.split("/")[-1]
#             if not files_s3_map.get(output_file):
#                 files_s3_map[output_file] = [s3_link_uuid, s3_object_url]

#             file.close()
#             s3_link_ids.append(s3_link_uuid)
#             s3_object_urls.append(s3_object_url)
#             kwargs["files_s3_map"] = files_s3_map

#         return {"s3_link_ids": s3_link_ids, "s3_object_urls": s3_object_urls}
#     except Exception as e:
#         if kwargs.get("tracking_ids"):
#             failed_tracking_ids_key = f'failed_tracking_id_{kwargs.get("batch_id","")}'
#             redis_instance.lpush(failed_tracking_ids_key, *kwargs.get("tracking_ids"))
#             redis_instance.close()
#         logger.error(f"upload_notices_to_s3.upload_failure : {str(e)}")
#         raise Exception(f"upload_notices_to_s3.exception :: {str(e)}")


# def gen_short_key(length=6):
#     valid_chars = string.ascii_letters + string.digits
#     short_key = ""
#     for i in range(length):
#         short_key += random.SystemRandom().choice(valid_chars)
#     return short_key


# def insert_into_db(**kwargs):
#     logger.info("notice.insert_into_db")
#     logger.info(f"notice.insert_into_db.batch_id :: {kwargs['batch_id']}")
#     loan_allocation_map = kwargs["loan_allocation_map"]
#     is_linked_loan = True if kwargs.get("is_linked_loan") else False
#     is_in_case = False

#     if kwargs.get("is_in_case", ""):
#         is_in_case = True

#     if "physical" in kwargs["mode"].lower():
#         notice_mode = "physical"
#         kwargs["document_type"] = "Notice"
#     else:
#         notice_mode = "digital"

#     data_to_dump = []

#     if notice_mode == "digital":
#         if isinstance(kwargs["loan_ids"], str):
#             loan_ids = kwargs["loan_ids"].split(", ")
#         else:
#             loan_ids = kwargs["loan_ids"]

#         s3_link_ids = kwargs.get("s3_link_ids", [])
#         s3_object_urls = kwargs.get("s3_object_urls", [])

#         for loan_id in loan_ids:
#             for i in range(len(s3_link_ids)):
#                 data_dict = {}
#                 data_dict["loan_id"] = loan_id
#                 data_dict["notice_type"] = kwargs.get("notice_type", "")
#                 data_dict["document_type"] = kwargs.get("document_type", "")
#                 data_dict["s3_link"] = s3_object_urls[i]
#                 data_dict["s3_link_uuid"] = s3_link_ids[i]
#                 data_dict["data"] = kwargs.get("db_data", {})
#                 data_dict["author"] = kwargs.get("author", "")
#                 data_dict["role"] = kwargs.get("role", "")
#                 data_dict["allocation_month"] = loan_allocation_map[data_dict["loan_id"]]
#                 data_dict["case_type"] = kwargs.get("case_type", "")
#                 data_dict["is_in_case"] = is_in_case
#                 data_dict["case_id"] = kwargs.get("case_id", "")
#                 data_dict["stage_code"] = kwargs.get("stage_code", 0)
#                 data_dict["iteration"] = kwargs.get("iteration", 0)
#                 data_dict["is_linked_loan"] = is_linked_loan
#                 data_dict["linked_loan_id"] = kwargs.get("linked_loan_id", "")
#                 data_dict["batch_id"] = kwargs.get("batch_id", "")
#                 data_dict["notice_mode"] = notice_mode
#                 data_dict["company_id"] = kwargs.get("company_id", "")
#                 data_dict["created"] = str(datetime.now())
#                 data_dict["is_dsc_signed"] = kwargs.get("is_dsc_signed", False)
#                 data_dict["dsc_placement"] = kwargs.get("dsc_placement")
#                 data_dict["primary_notice_data"] = kwargs["primary_notice_data"]
#                 data_to_dump.append(data_dict)
#     else:
#         loan_address_file_s3_map_list = kwargs["loan_address_file_s3_map_list"]
#         for row in loan_address_file_s3_map_list:
#             row["company_id"] = kwargs.get("company_id", "")
#             row["notice_type"] = kwargs.get("notice_type", "")
#             row["document_type"] = kwargs.get("document_type")
#             row["data"] = copy.deepcopy(kwargs.get("db_data", {}))
#             row["author"] = kwargs.get("author", "")
#             row["role"] = kwargs.get("role")
#             row["allocation_month"] = loan_allocation_map[row["loan_id"]]
#             row["case_type"] = kwargs.get("case_type")
#             row["is_in_case"] = kwargs.get("is_in_case", False)
#             row["case_id"] = kwargs.get("case_id", "")
#             row["stage_code"] = kwargs.get("stage_code", 0)
#             row["iteration"] = kwargs.get("iteration", 0)
#             row["is_linked_loan"] = is_linked_loan
#             row["linked_loan_id"] = kwargs.get("linked_loan_id")
#             row["batch_id"] = kwargs["batch_id"]
#             row["tracking_ids"] = kwargs.get("tracking_ids", [])
#             row["notice_mode"] = notice_mode
#             row["primary_notice_data"] = kwargs["primary_notice_data"]
#             if row.get("tracking_id"):
#                 row["notice_tracking_data"] = get_notice_tracking_data(row, **kwargs)
#         data_to_dump = loan_address_file_s3_map_list

#         for data in data_to_dump:
#             data["data"]["borrower_type"] = data["borrower_type"]

#     try:
#         url = f"{NOTICE_SERVICE_BASE_URL.rstrip('/')}/v2/notices"
#         headers = {
#             "authenticationtoken": kwargs.get("auth_token", ""),
#             "role": kwargs.get("role", ""),
#             "Content-type": "application/json",
#             "X-Request-Id": kwargs.get("X-Request-Id", ""),
#             "X-CG-User": json.dumps(kwargs.get("user", {})),
#             "X-CG-Company": json.dumps(kwargs.get("company", {})),
#         }

#         insert_result_resp = requests.post(
#             url=url,
#             json={"payload": data_to_dump, "company_id": kwargs["company_id"]},
#             data=json.dumps({"payload": data_to_dump, "company_id": kwargs["company_id"]}),
#             headers=headers,
#         )

#         logger.debug(f"notice.insert_into_db.url : {url}")
#         logger.debug(f"notice.insert_into_db.response.status_code : {insert_result_resp.status_code}")
#         insert_result = json.loads(insert_result_resp.text)
#         if insert_result.get("message") != "SUCCESS":
#             raise Exception(f"{insert_result.get('data')}")

#     except Exception as e:
#         # delete the notices from EFS in case of db insertion failure
#         if notice_mode == "physical":
#             generated_notice_pdfs_path = kwargs["local_pdf_directory_path"]
#             pdf_files_to_remove = [
#                 os.path.join(generated_notice_pdfs_path, f"{notice_id}.pdf")
#                 for notice_id in kwargs.get("notice_id_list", [])
#             ]
#             generated_notice_docs_path = kwargs["local_doc_directory_path"]
#             docx_files_to_remove = [
#                 os.path.join(generated_notice_docs_path, f"{notice_id}.docx")
#                 for notice_id in kwargs.get("notice_id_list", [])
#             ]

#             for file in docx_files_to_remove:
#                 os.remove(file)
#             for file in pdf_files_to_remove:
#                 os.remove(file)

#         logger.error(f"notice.insert_into_db.exception : {str(e)}")
#         raise Exception(f"notice.insert_into_db.exception : {str(e)}")


# def get_notice_tracking_data(notice_details, **kwargs):
#     logger.info("notice.get_notice_tracking_data")
#     notice_address_data = json.loads(notice_details["primary_address"])

#     speedpost_metadata = {
#         "speedpost_id": notice_details.get("tracking_id", ""),
#         "speedpost_tarrif": "",
#         "booked_at": "",
#         "delivery_location": "",
#         "article_type": "",
#         "speedpost_tariff": "",
#         "speedpost_booked_on": "",
#         "speedpost_delivery_status": "In-transit",
#         "speedpost_undelivered_reason": "",
#         "customer_name": notice_address_data.get("table_1_cell_1_name", ""),
#         "recipient_type": notice_address_data.get("table_1_borrower_type", ""),
#         "field_type": "",
#         "customer_address": notice_address_data.get("table_1_address_text", ""),
#         "customer_pincode": notice_address_data.get("table_1_pincode", ""),
#         "speedpost_destination_pincode": notice_address_data.get("table_1_pincode", ""),
#         "speedpost_delivery_confirmed_on": "",
#         "applicant_type": notice_address_data.get("table_1_borrower_type", ""),
#         "address_type": notice_details["address_type"],
#         "applicant_index": "",
#         "address_index": "",
#         "city": notice_address_data.get("table_1_city", ""),
#         "state": notice_address_data.get("table_1_state", ""),
#         "notice_reference_number": notice_details.get("data", {}).get("notice_reference_number"),
#         "applicant_notice_reference_number": notice_details.get("data", {}).get("applicant_notice_reference_number"),
#         "co_applicant_notice_reference_number": notice_details.get("data", {}).get(
#             "co_applicant_notice_reference_number"
#         ),
#         "events": [],
#         "speedpost_upload_datetime": str(datetime.now()),
#     }

#     speedpost_db_data = {
#         "company_id": notice_details.get("company_id", ""),
#         "loan_id": notice_details.get("loan_id", ""),
#         "notice_type": notice_details.get("notice_type", ""),
#         "s3_link": notice_details.get("s3_link", ""),
#         "s3_link_uuid": notice_details.get("s3_link_uuid", ""),
#         "document_type": "Speedpost",
#         "data": json.dumps(speedpost_metadata),
#         "allocation_month": notice_details.get("allocation_month", ""),
#         "notice_batch_id": kwargs["batch_id"],
#         "tracking_id": notice_details["tracking_id"],
#     }

#     return speedpost_db_data



# import json
# import logging
# import os
# import uuid

# import redis
# from celery import shared_task

# from app.tasks.notice.loan_service import get_loan_data
# from app.tasks.notice.notice_service import (
#     get_notice_generation_data,
#     insert_into_db,
#     upload_notices_to_s3,
# )

# from ...settings import (
#     NOTICE_SERVICE_BASE_URL,
#     RECOVERY_SERVICE_BASE_URL,
#     REDIS,
#     S3_BUCKET_ENDPOINT,
#     S3_BUCKET_NAME,
# )
# from .helpers import (
#     doc2pdf_converter,
#     generate_qrcode_doc,
#     get_dsc_on_notice_pdf,
#     logs_tracker,
#     make_notice_doc_file,
# )

# logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# @shared_task(
#     bind=True,
#     name="handle_generate_digital_notice",
#     ignore_result=False,
#     acks_late=False,
# )
# def handle_generate_digital_notice(self, *args, **kwargs):
#     logger.info("notice.handle_generate_digital_notice")
#     is_task_success = False

#     logs_loan_id = kwargs["linked_loan"]["linked_loan_id"] if kwargs.get("linked_loan") else kwargs.get("loan_ids")[0]
#     logger.info(f"notice.handle_generate_digital_notice.kwargs :: {kwargs}")
#     logger.info(f"notice.handle_generate_digital_notice.loan_id :: {logs_loan_id}")
#     logger.info(f"notice.handle_generate_digital_notice.batch_id :: {kwargs['batch_id']}")

#     try:
#         redis_instance = redis.Redis(host=REDIS["HOST"], port=REDIS["PORT"])
#         notice_kwargs = json.loads(redis_instance.get(f"notice_kwargs_{kwargs['batch_id']}").decode("utf-8"))
#         redis_instance.close()
#         # In case key is not present in redis
#         if not notice_kwargs:
#             logger.error("notice.handle_generate_digital_notice.exception: notice_kwargs not found in redis")
#             raise Exception("notice_kwargs not found in redis")

#         kwargs = {**kwargs, **notice_kwargs}
#         logger.debug(f"notice.generate_digital_notices.kwargs : {kwargs}")

#         local_batch_directory_path = kwargs["local_batch_directory_path"]
#         local_draft_file_path = kwargs["local_draft_file_path"]
#         local_doc_directory_path = kwargs["local_doc_directory_path"]
#         local_pdf_directory_path = kwargs["local_pdf_directory_path"]
#         local_qr_code_directory_path = kwargs["local_qr_code_directory_path"]
#         preview = kwargs.get("preview", False)
#         request_id = str(uuid.uuid4())

#         os.makedirs(local_doc_directory_path, exist_ok=True)
#         os.makedirs(local_pdf_directory_path, exist_ok=True)
#         os.makedirs(local_qr_code_directory_path, exist_ok=True)

#         # For linked loans
#         if kwargs.get("linked_loan"):
#             linked_loan_dict = kwargs["linked_loan"]
#             linked_loan_id = linked_loan_dict.get("linked_loan_id", "")
#             loan_ids = linked_loan_dict.get("loan_ids", [])

#             if linked_loan_id in (None, "") or len(loan_ids) == 0:
#                 raise Exception(f"Invalid linked loan id {linked_loan_id}")
#         # For normal loans
#         else:
#             loan_ids = kwargs["loan_ids"]

#         notice_generation_kwargs = {
#             "company_id": kwargs["company_id"],
#             "draft_id": kwargs["draft_id"],
#             "draft_details": kwargs["draft_details"],
#             "loan_ids": loan_ids,
#             "linked_loan_id": linked_loan_id if kwargs.get("linked_loan") else "",
#             "allocation_month": kwargs["allocation_month"],
#             "author": kwargs["author"],
#             "role": kwargs["role"],
#             "batch_number": kwargs["batch_number"],
#             "batch_id": kwargs["batch_id"],
#             "document_type": kwargs["document_type"],
#             "mode": kwargs["mode"],
#             "type_of_task": "digital",
#             "preview": preview,
#             "auth_token": kwargs["token"],
#             "BUCKET_NAME": S3_BUCKET_NAME,
#             "S3_BUCKET_ENDPOINT": S3_BUCKET_ENDPOINT,
#             "RECOVERY_SERVICE_BASE_URL": RECOVERY_SERVICE_BASE_URL,
#             "local_pdf_directory_path": local_pdf_directory_path,
#             "local_doc_directory_path": local_doc_directory_path,
#             "local_batch_directory_path": local_batch_directory_path,
#             "local_draft_file_path": local_draft_file_path,
#             "local_qr_code_directory_path": local_qr_code_directory_path,
#             "local_merged_pdf_directory_path": kwargs["local_merged_pdf_directory_path"],
#             "local_merged_pdf_zip_file_path": kwargs["local_merged_pdf_zip_file_path"],
#             "NOTICE_SERVICE_BASE_URL": NOTICE_SERVICE_BASE_URL,
#             "user": kwargs["user"],
#             "is_linked_loan": kwargs["is_linked_loan"],
#             "case_type": kwargs.get("case_type"),
#             "case_id": kwargs.get("case_id"),
#             "iteration": kwargs.get("iteration"),
#             "stage_code": kwargs.get("stage_code"),
#             "request_id": request_id,
#             "draft_fields_redis_key": kwargs["draft_fields_redis_key"],
#             "company": kwargs["company"],
#             "enable_qrcode": kwargs["enable_qrcode"],
#         }

#         if preview:
#             notice_generation_kwargs["s3_link_uuid"] = kwargs["s3_link_uuid"]

#         generate_digital_notice(**notice_generation_kwargs)

#         message = "Linked loan digital notice generated" if kwargs.get("linked_loan_id") else "Digital notice generated"

#         logs_tracker_req_body = {
#             "response": {"message": f"{message}"},
#             "kwargs": {
#                 "batch_id": kwargs["batch_id"],
#                 "author": kwargs["user"]["email"],
#                 "company_id": kwargs["company"]["company_id"],
#                 "loan_id": logs_loan_id,
#                 "user": kwargs["user"],
#                 "company": kwargs["company"],
#             },
#             "status_code": 200,
#             "status": "SUCCESS",
#         }
#     except Exception as e:
#         logger.error(f"notice.handle_generate_digital_notice.exception: {str(e)}")

#         logs_tracker_req_body = {
#             "response": str({"message": str(e)}),
#             "kwargs": {
#                 "batch_id": kwargs["batch_id"],
#                 "batch_number": kwargs["batch_number"],
#                 "total_batches": 1,
#                 "author": kwargs.get("author"),
#                 "company_id": kwargs.get("company_id"),
#                 "loan_id": logs_loan_id,
#                 "total_loans": 1,
#                 "user": kwargs.get("user", {}),
#                 "company": kwargs.get("company", {}),
#             },
#             "status_code": 400,
#             "status": "FAIL",
#         }
#     finally:
#         logs_tracker(logs_tracker_req_body, kwargs.get("X-Request-Id"))
#         return is_task_success


# def generate_digital_notice(**kwargs):
#     logger.info("tasks.notice.generate_digital_notice")
#     logger.info(f"notice.handle_generate_digital_notice.batch_id :: {kwargs['batch_id']}")

#     try:
#         kwargs["X-Request-Id"] = kwargs["request_id"]
#         local_pdf_directory_path = kwargs.get("local_pdf_directory_path")
#         redis_instance = redis.Redis(host=REDIS["HOST"], port=REDIS["PORT"])

#         draft_fields = redis_instance.get(kwargs["draft_fields_redis_key"])
#         if not draft_fields:
#             logger.error("tasks.generate_digital_notice.exception: Unable to fetch draft details from redis")
#             raise Exception("tasks.generate_digital_notice.exception: Unable to fetch draft details from redis")

#         draft_fields = json.loads(draft_fields.decode("utf-8"))
#         redis_instance.close()
#         preview = kwargs.get("preview", False)
#         draft_details = kwargs.get("draft_details", {})
#         kwargs["draft_name"] = draft_details.get("draft_name")
#         kwargs["notice_type"] = draft_details.get("notice_type")
#         kwargs["is_vernacular"] = draft_details.get("is_vernacular")
#         kwargs["is_in_case"] = draft_fields.get("is_in_case")
#         kwargs["language"] = draft_details.get("language")
#         kwargs["draft_fields"] = draft_fields.get(kwargs.get("draft_id"))
#         kwargs["is_dsc_enabled"] = draft_details.get("is_dsc_enabled")
#         kwargs["dsc_placement"] = draft_details.get("dsc_placement")

#         loan_data = get_loan_data(**kwargs)

#         if not loan_data:
#             logger.error("generate_digital_notice.get_loan_data_res.err : Can't find loan data")
#             raise Exception("Cannot find data for the given loan")
#         if isinstance(loan_data, str):
#             logger.error(f"generate_digital_notice.get_loan_data_res.err : {loan_data}")
#             raise Exception(loan_data)

#         if len(kwargs.get("loan_ids", [])) != len(loan_data["company_data"]):
#             logger.error(f"generate_digital_notice.get_loan_data_res.err : Can't find data for all the loans")
#             raise Exception("Can't find data for all the loans")

#         kwargs["loan_data"] = loan_data["company_data"]
#         kwargs["loan_address_map"] = {}
#         notice_data = get_notice_generation_data(**kwargs)

#         kwargs["filenames"] = []
#         kwargs["index"] = 1

#         languages = notice_data.get("languages", {})
#         kwargs["notice_data"] = notice_data
#         master_table_dict = notice_data["master_table_dict"]
#         duplicated_tables_dict = notice_data["duplicated_tables_dict"]
#         kwargs["notice_data"]["master_table_dict"] = {
#             **master_table_dict,
#             **duplicated_tables_dict,
#         }

#         notice_file_data = make_notice_doc_file(**kwargs)

#         input_file, output_file = (
#             notice_file_data["input_file_path"],
#             notice_file_data["output_file_path"],
#         )
#         kwargs["input_file"] = input_file
#         kwargs["output_file"] = output_file
#         output_folder = local_pdf_directory_path

#         if kwargs.get("enable_qrcode", False):
#             generate_qrcode_doc(**kwargs)

#         doc2pdf_converter(output_folder, **kwargs)

#         if kwargs["is_dsc_enabled"]:
#             get_dsc_on_notice_pdf(**kwargs)
#             kwargs["is_dsc_signed"] = True

#         kwargs["filenames"].append(output_file)
#         changed_dt = notice_data.get("changed_dt", {})

#         db_data = {
#             "barcode": changed_dt.get("barcode"),
#             "applicant_type": changed_dt["applicant_type"],
#             "borrower_type": changed_dt["applicant_type"],
#             "address_index": changed_dt["address_index"],
#             "batch_number": kwargs.get("batch_number", ""),
#             "notice_reference_number": changed_dt["notice_reference_number"],
#             "applicant_notice_reference_number": changed_dt["applicant_notice_reference_number"],
#             "co_applicant_notice_reference_number": changed_dt["co_applicant_notice_reference_number"],
#             "applicant_language": languages.get("applicant_language"),
#             "co_applicant_language": languages.get("co_applicant_language"),
#             "draft_language": languages.get("draft_language"),
#             "notice_variables_data": changed_dt,
#         }

#         kwargs["db_data"] = db_data
#         upload_response = upload_notices_to_s3(**kwargs)

#         kwargs["loan_allocation_map"] = get_loan_allocation_month_mapping(kwargs["loan_data"])

#         if os.path.exists(input_file):
#             os.remove(input_file)
#         if os.path.exists(output_file):
#             os.remove(output_file)

#         if not preview:
#             kwargs["s3_link_ids"] = upload_response["s3_link_ids"]
#             kwargs["s3_object_urls"] = upload_response["s3_object_urls"]
#             kwargs["primary_notice_data"] = {
#                 "s3_link_uuid": upload_response["s3_link_ids"][0],
#                 "s3_link": upload_response["s3_object_urls"][0],
#             }
#             insert_into_db(**kwargs)

#     except Exception as e:
#         raise Exception(f"tasks.notice.generate_digital_notice.exception :: {str(e)}")


# def get_loan_allocation_month_mapping(loan_data):
#     loan_allocation_map = {}
#     logger.info("generate_physical_notice.get_loan_allocation_month_mapping")
#     for loan in loan_data:
#         loan_id = loan["loan_id"]
#         allocation_month = loan["question_dict"]["allocation_month"]
#         loan_allocation_map[loan_id] = allocation_month
#     return loan_allocation_map
